from jinja2 import Environment
from jinja2.utils import contextfunction


def _get_authx_config(context, name):
    return context.parent.get('authx', {}).get(name, {})


def has_ipaccess(context, name=None):
    config = _get_authx_config(context, 'ipaccess')
    if len(config) > 0 and (name is None or name in config):
        return True
    return False


def has_token_attr(context, key, value):
    config = _get_authx_config(context, 'bearer')
    if value == config.get(key, ""):
        return True
    return False


def is_bearer_auth(context, provider=None):
    config = _get_authx_config(context, 'bearer')
    if len(config) > 0 and (provider is None or provider == config['__provider_name']):
        return True
    return False


def has_basic_auth(context):
    config = _get_authx_config(context, 'basic')
    if len(config) > 0:
        return True
    return False


def is_basic_user(context, user=None):
    config = _get_authx_config(context, 'basic')
    if len(config) > 0 and (user is None or config.get('username') == user):
        return True
    return False


def has_basic_group(context, group):
    config = _get_authx_config(context, 'basic')
    if group in config.get('groups', []):
        return True
    return False


def has_cookie(context, name=None):
    config = _get_authx_config(context, 'cookie')
    if len(config) > 0 and (name is None or name in config):
        return True
    return False


def has_any_auth(context):
    if len(context.parent.get('authx', {})) > 0:
        return True
    return False


_environment = Environment()
_environment.globals.update(
    {
        'has_ipaccess': contextfunction(has_ipaccess),
        'is_bearer_auth': contextfunction(is_bearer_auth),
        'has_token_attr': contextfunction(has_token_attr),
        'has_basic_auth': contextfunction(has_basic_auth),
        'is_basic_user': contextfunction(is_basic_user),
        'has_basic_group': contextfunction(has_basic_group),
        'has_cookie': contextfunction(has_cookie),
        'has_any_auth': contextfunction(has_any_auth)
    }
)


# TODO: @lru_cache
def validator(condition="False", config={}, **kwargs) -> bool:
    template = _environment.from_string(f"{{{{ {condition} }}}}")
    result = template.render(authx=config, call_env=kwargs)
    return False if result != 'True' else True




class InvalidToken(Exception):
    pass


def dict_merge(base_dct, merge_dct):
    base_dct.update({
        key: dict_merge(base_dct[key], merge_dct[key])
        if isinstance(base_dct.get(key), dict) and isinstance(merge_dct[key], dict)
        else merge_dct[key]
        for key in merge_dct.keys()
    })


def merge_auth_info(scope, data):
    if data is not None:
        if 'authx' not in scope:
            scope['authx'] = {}
        dict_merge(scope['authx'], data)



import json
import urllib.request
import jose.jwt
import logging
from starlette.requests import Request
from starlette.types import Receive, Scope, Send
from . import InvalidToken, merge_auth_info

_jwt_keys = {}

"""
this stuff is mostly from https://gitlab.com/jorgecarleitao/starlette-oauth2-api
"""


def _get_keys(url_or_keys):
    if isinstance(url_or_keys, str) and url_or_keys.startswith('https://'):
        with urllib.request.urlopen(url_or_keys) as f:
            return json.loads(f.read().decode())
    else:
        return url_or_keys


async def process(config, scope: Scope, receive: Receive, send: Send) -> None:
    """
    this stuff is mostly from https://gitlab.com/jorgecarleitao/starlette-oauth2-api
    """
    client_bearer_auth_results = {}
    request = Request(scope)

    # check for authorization header and token on it.
    if 'authorization' in request.headers and request.headers['authorization'].startswith('Bearer '):
        token = request.headers['authorization'][len('Bearer '):]
        providers = config.get('providers')
        for provider in providers:
            try:
                result = _provider_claims(provider, providers[provider], token)
                if len(client_bearer_auth_results) == 1:
                    logging.warning(f"more than one provider matches that token. looks like a config error.")
                else:
                    client_bearer_auth_results = result
                    client_bearer_auth_results['__provider_name'] = provider
            except jose.exceptions.JWSError as e:
                raise InvalidToken(str(e))
            except jose.exceptions.JWTClaimsError as e:
                pass
            except jose.exceptions.JWTError as e:
                raise InvalidToken(str(e))
            except jose.exceptions.JOSEError as e:
                logging.debug(str(e))

    if len(client_bearer_auth_results) > 0:
        merge_auth_info(scope, {'bearer': client_bearer_auth_results})


def _provider_claims(provider_name, provider, token):
    """

    """
    return jose.jwt.decode(
        token, _provider_keys(provider_name, provider),
        issuer=provider['issuer'],
        audience=provider['audience'],
        options={'verify_at_hash': False}
    )


def _provider_keys(provider_name, provider):
    """
    Returns the signing keys of the provider
    """
    if _jwt_keys.get(provider_name, None) is None:
        _jwt_keys[provider_name] = _get_keys(provider['keys'])
    return _jwt_keys[provider_name]



import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), 'r') as fh:
    long_description = fh.read()


setuptools.setup(
    name='starlette-authx',
    version='0.0.22',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Christoph Hartmann',
    author_email='mail_to_chriss@posteo.net',
    url='https://github.com/chriss-de/starlette-authx',
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'starlette>=0.12,<1',
        'python-jose>=3,<4',
        'passlib>=1,<2',
        'jinja2>=2,<3',
        'itsdangerous>=1.1.0,<2',
        'requests-oauthlib>=1.3.0,<2',
    ],
    extras_require={
        'gql': ['graphene>2.1,<3'],
        'fastapi': ['fastapi>=0.54,<1'],
    },
)



from starlette.types import Receive, Scope, Send
import ipaddress
from . import merge_auth_info


async def process(config, scope: Scope, receive: Receive, send: Send) -> None:
    client_networks = []
    client_ip_address = ipaddress.ip_address(scope['client'][0])

    for ip_access in config:
        for ip_address in config[ip_access]:
            if client_ip_address in ipaddress.ip_interface(ip_address).network:
                client_networks.append(ip_access)

    if len(client_networks) > 0:
        merge_auth_info(scope, {'ipaccess': set(client_networks)})
